
#include <stdlib.h>
#include <fstream>
#include <irdb-core>
#include <libgen.h>
#include <assert.h>

#include <stars.h>

using namespace std;
using namespace IRDB_SDK;

/* include the interface to STARS */
#include "interfaces/abstract/STARSInterface.h"
#include "interfaces/irdb/STARSInterface.h"

/* include the STARS def. of a program.  -- aka a file */
#include "interfaces/abstract/STARSProgram.h"
#include "interfaces/irdb/STARSProgram.h"

#include "interfaces/irdb/STARSInstruction.h"



/* include STARS data structures. */
#include "SMPProgram.h"
#include "base/ProfilerInformation.h"

STARS_Interface_t* global_stars_interface = NULL;
STARS_Program_t *global_STARS_program = NULL;


void Init_stars_global_program(FileIR_t *firp)
{
	assert(firp);

	// set 32 or 64bit-ness of program
	if(firp->getArchitectureBitWidth()==64)
		// note: irdb version currently exists, but
		// it's going away and code to move to base class
		// clc doing this.
		global_STARS_program->Set64BitBinary();		
	else
		global_STARS_program->Set32BitBinary();		


	// note: irdb version currently exists, but
	// it's going away and code to move to base class
	// clc doing this.
	global_STARS_program->InitData();

	// this should work right now.
	// no path --  just the base filename, could pass "a.ncexe" by def.
	string filename=firp->getFile()->getURL();

	global_STARS_program->SetRootFileName( filename.substr(filename.find_last_of("/\\")+1) );

	//  no other bits needed.
	global_STARS_program->OpenFiles();

	//  read_zephyr_security_toolkit policies	
	global_STARS_program->ZST_InitPolicies();

}

SMPProgram* InitSMPProgram()
{
	// start creating the STARS data structures!

	// allocate program.
	SMPProgram* cur_pgm=new SMPProgram;

	// call may be a nop, but leave it in case we do stuff later.
	cur_pgm->AnalyzeData();

	// FIXME: sometime in the future, do the ProfilerInformation read-in	
	// skip for now.

	// signal that curr. Profiler info processing is complete. 
	cur_pgm->ProfGranularityFinished( global_STARS_program->GetAnnotFile(), 
		global_STARS_program->GetInfoAnnotFile());
	return cur_pgm;

}


void ProcessOrphanedInstrs(SMPProgram*  cur_pgm, FileIR_t* firp)
{
	set<STARS_IRDB_Instruction_t*> orphans;

	STARS_Program_t* gsp=global_STARS_program;
	// for each orphan instruction (i.e., instruction that's not part of a function
	for(auto it=firp->getInstructions().begin();
		it!=firp->getInstructions().end();
		++it)
	{
		Instruction_t* instr=*it;
		assert(instr);
		assert(instr->getBaseID() != BaseObj_t::NOT_IN_DATABASE);

		if(instr->getFunction()!=NULL)
			continue;

		// create the IRDB instruction
		STARS_IRDB_Instruction_t *foo= new STARS_IRDB_Instruction_t(instr);
		orphans.insert(foo);


		// instr is "orphaned", that is, not part of a function.

		// create the SMPInstr analyze it.
		SMPInstr CurrInst(instr->getBaseID());
		CurrInst.Analyze();
	
		// Do machine-dependent fixes for DEF and USE lists.
		//  The fixes can help produce better annotations.
		CurrInst.MDFixupDefUseLists();
	
		// If instruction is still not included in a code chunk,
		//  emit annotations for it in isolation.
		if (CurrInst.IsAnalyzeable()) 
		{
			CurrInst.EmitAnnotations(true, false, true, 
				gsp->GetAnnotFile(), gsp->GetInfoAnnotFile(), 
				cur_pgm);
		
		}


		// delete the insn here?
		delete foo;
	}

	// or process orphans and delete here?
	// for ( each item orphan) delete item;

}


void OutputStatistics()
{
#if 0
		// Output statistics, not necessary yet.

		for (int OptType = 0; OptType <= LAST_OPT_CATEGORY; ++OptType) {
				SMP_msg("Optimization Category Count %d:  %d Annotations: %d\n",
						OptType, global_STARS_program->GetOptCount(OptType), global_STARS_program->GetAnnotationCount(OptType));
		}
#endif
	// mark that we've completed analysis.	
	SMP_fprintf(global_STARS_program->GetInfoAnnotFile(), "   8000000      2 SUCCESS ANALYSISCOMPLETED\n");
	STARS_Instruction_t::DumpCreationStats(cout);
	STARS_op_t::DumpCreationStats(cout);
}

int STARS::IRDB_Interface_t::do_STARS(FileIR_t* firp)
{
	STARS_InstructionID_t::Init();
	STARS_IRDB_Program_t* gsp=NULL;
	global_stars_interface = new STARS_IRDB_Interface_t(firp, pqxx_interface);
	global_STARS_program = gsp = new STARS_IRDB_Program_t(firp);

	//  this is initing the global_stars_program (GSP) 


	// do global initialization of GSP
	Init_stars_global_program(firp);

	// Set options into the GSP
	if (this->GetSTARSOptions().ShouldSTARSPerformLevel2ReducedAnalysis()) {
		global_STARS_program->SetLevel2ReducedProcessingFlag(true);
		SMP_msg("INFO: Level 2 Reduced processing requested via command-line options.\n");
	}
	else if (this->GetSTARSOptions().ShouldSTARSPerformReducedAnalysis()) {
		global_STARS_program->SetReducedProcessingFlag(true);
		SMP_msg("INFO: Reduced processing requested via command-line options.\n");
	}
	if (this->GetSTARSOptions().ShouldSTARSShadowFuncPtrs()) {
		global_STARS_program->SetFuncPtrShadowingFlag(true);
		SMP_msg("INFO: Function pointer shadowing requested via command-line options.\n");
	}
	if (this->GetSTARSOptions().ShouldSTARSShadowCriticalArgs()) {
		global_STARS_program->SetArgShadowingFlag(true);
		SMP_msg("INFO: Critical argument shadowing requested via command-line options.\n");
	}
	if (this->GetSTARSOptions().ShouldSTARSMaximizeCFGImprovement()) {
		global_STARS_program->SetCFGImprovementFlag(true);
		SMP_msg("INFO: Maximum CFG improvement requested via command-line options.\n");
	}
	if (this->GetSTARSOptions().ShouldSTARSPerformDeepLoopAnalyses()) {
		global_STARS_program->SetLoopAnalysesFlag(true);
		SMP_msg("INFO: Deep loop analyses requested via command-line options.\n");
	}
	if (this->GetSTARSOptions().ShouldSTARSPerformConstantPropagation()) {
		global_STARS_program->SetConstantPropagationFlag(true);
		SMP_msg("INFO: Constant propagation via SCCP requested via command-line options.\n");
	}

	// create and init a SMP program.
	SMPProgram* cur_pgm = InitSMPProgram();

	// setup profiler info -- to do later.  FIXME.
	ProfilerInformation* profinfo=new ProfilerInformation(global_STARS_program->GetAnnotFileName().c_str(),cur_pgm);

	// analyze the program, as in, invoke stars.
	cur_pgm->Analyze(profinfo, gsp->GetAnnotFile(), gsp->GetInfoAnnotFile());




	// emit annotations.
	if (!global_STARS_program->ShouldSTARSPerformReducedAnalysis()) {
			cur_pgm->EmitAnnotations(
		gsp->GetAnnotFile(), gsp->GetInfoAnnotFile());
	}

	// deal with annotations for instructions that aren't part of a function.
	ProcessOrphanedInstrs( cur_pgm, firp );

	// emit any stats.	
	OutputStatistics();

	// cleanup open files.
	global_STARS_program->CloseFiles();


	// now re-read them with a meds annotation parser
	parser.parseFile(gsp->GetAnnotFileName());
	parser.parseFile(gsp->GetInfoAnnotFileName());

	// cleanup memory allocated by driver.
	delete profinfo;
	profinfo=NULL;

	delete cur_pgm;
	cur_pgm=NULL;

	delete global_stars_interface; 
	global_stars_interface=NULL;

	delete global_STARS_program; 
	global_STARS_program=NULL;

	return 1;
}

